/*
 *
 * Copyright (C) 2008 Ivo Sedlacek

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

 * Functions delay_1ms, adc_measure, get_pin_adc, set_dir, send_msg, refsms,
 *    refsms_d, pager_clear_sim_sms and function fragments in ThermostatReceiveCall
 *    and ThermostatReceiveSMS - Copyright (C) 2004, 2005, 2006 BLADOX, s.r.o.
 *
 */

#include <config.h>
#include <turbo/turbo.h>

#include <avr/sfr_defs.h>
#include <avr/io.h>

#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#define WITH_REMOTE_NUMBERS
//#define SMS_WITH_LOGGING

/* *INDENT-OFF* */

/* strings for menu */
lc_char PROGMEM lc_Thermostat[] = {
  LC_EN ("Thermostat")
  LC_CZ ("Termostat")
  LC_END
};

lc_char PROGMEM lc_State[] = {
  LC_EN ("State")
  LC_CZ ("Stav")
  LC_END
};

lc_char PROGMEM lc_StateOn[] = {
  LC_EN ("On")
  LC_CZ ("Zapnuto")
  LC_END
};

lc_char PROGMEM lc_StateOff[] = {
  LC_EN ("Off")
  LC_CZ ("Vypnuto")
  LC_END
};

lc_char PROGMEM lc_TargetTemperature[] = {
  LC_EN (", set to ")
  LC_CZ (", nastaveno na ")
  LC_END
};

lc_char PROGMEM lc_CurrentTemperature[] = {
  LC_EN (" C, currently ")
  LC_CZ (" C, aktualne ")
  LC_END
};

lc_char PROGMEM lc_Celsius[] = {
  LC_EN (" C")
  LC_END
};

#ifdef WITH_REMOTE_NUMBERS
lc_char PROGMEM lc_RemoteNumbersStart[] = {
  LC_EN (", control (")
  LC_CZ (", ovladani (")
  LC_END
};

lc_char PROGMEM lc_RemoteNumbersMiddle[] = {
  LC_EN (", ")
  LC_END
};

lc_char PROGMEM lc_OpenBracket[] = {
  LC_EN ("(")
  LC_END
};

lc_char PROGMEM lc_Pointer[] = {
  LC_EN (" to ")
  LC_END
};

lc_char PROGMEM lc_CloseBracket[] = {
  LC_EN (")")
  LC_END
};

lc_char PROGMEM lc_RemoteNumbersEnd[] = {
  LC_EN (")")
  LC_END
};
#endif

lc_char PROGMEM lc_Setting[] = {
  LC_EN ("Setting")
  LC_CZ ("Nastaveni")
  LC_END
};

lc_char PROGMEM lc_Turn[] = {
  LC_EN ("On/Off")
  LC_CZ ("Zap/Vyp")
  LC_END
};

lc_char PROGMEM lc_Off[] = {
  LC_EN ("  Off")
  LC_CZ ("  Vypnout")
  LC_END
};

lc_char PROGMEM lc_On[] = {
  LC_EN ("  On")
  LC_CZ ("  Zapnout")
  LC_END
};

lc_char PROGMEM lc_OffWithStar[] = {
  LC_EN ("* Off")
  LC_CZ ("* Vypnuto")
  LC_END
};

lc_char PROGMEM lc_OnWithStar[] = {
  LC_EN ("* On")
  LC_CZ ("* Zapnuto")
  LC_END
};

lc_char PROGMEM lc_Temperature[] = {
  LC_EN ("Temperature")
  LC_CZ ("Teplota")
  LC_END
};

lc_char PROGMEM lc_ExpectedTemperature[] = {
  LC_EN ("Target temperature")
  LC_CZ ("Nastavena teplota")
  LC_END
};

lc_char PROGMEM lc_Numbers[] = {
  LC_EN ("Numbers")
  LC_CZ ("Cisla")
  LC_END
};

lc_char PROGMEM lc_Clear_SIM_SMS[]={
  LC_EN("Clear SIM SMS")
  LC_CZ("Smazat SIM SMS")
  LC_END
};

lc_char PROGMEM lc_Please_Wait[]={
  LC_EN("Please Wait")
  LC_CZ("Prosim cekejte")
  LC_END
};

lc_char PROGMEM lc_SMS_Deleted[]={
  LC_EN("All Messages on SIM Deleted")
  LC_CZ("Vsechny zpravy na SIM smazany")
  LC_END
};

lc_char PROGMEM lc_Answers[]={
  LC_EN("Answers to")
  LC_CZ("Odpovedi")
  LC_END
};

lc_char PROGMEM lc_ToSender[]={
  LC_EN("To sender's num.")
  LC_CZ("Odesilateli")
  LC_END
};

lc_char PROGMEM lc_Change[]={
  LC_EN("Change")
  LC_CZ("Zmenit")
  LC_END
};

/* *INDENT-ON* */


 /* spider node declaration */
extern SNodeP g_Thermostat;
extern SNodeP g_ThermostatState;
extern SNodeP g_ThermostatSetting;
extern SNodeP g_ThermostatSettingTurn;
extern SNodeP g_ThermostatSettingTurnOn;
extern SNodeP g_ThermostatSettingTurnOff;
extern SNodeP g_ThermostatSettingTurnOnWithStar;
extern SNodeP g_ThermostatSettingTurnOffWithStar;
extern SNodeP g_ThermostatSettingTemperature;
extern SNodeP g_ThermostatSettingNumbers;
extern SNodeP g_ThermostatSettingRemoveAllSms;
extern SNodeP g_ThermostatSettingNumbersNew;
extern SNodeP g_ThermostatSettingNumbersNumberDelete;
extern SNodeP g_ThermostatSettingNumbersNumberView;
extern SNodeP g_ThermostatSettingNumbersNumberAnswers;
extern SNodeP g_ThermostatSettingNumbersNumberAnswersNumView;
extern SNodeP g_ThermostatSettingNumbersNumberAnswersNumChange;
extern SNodeP g_ThermostatSettingNumbersNumberAnswersNumDelete;

/* spider callbacks */
extern u8 ThermostatStateCallBack(SCtx * ctx, u8 action);
extern u8 ThermostatSettingTurnCallBack(SCtx * ctx, u8 action);
extern u8 ThermostatSettingTurnOnCallBack(SCtx * ctx, u8 action);
extern u8 ThermostatSettingTurnOffCallBack(SCtx * ctx, u8 action);
extern u8 ThermostatSettingTemperatureCallBack(SCtx * ctx, u8 action);
extern u8 ThermostatSettingNumbersCallBack(SCtx * ctx, u8 action);
extern u8 ThermostatSettingNumbersNewCallBack(SCtx * ctx, u8 action);
extern u8 ThermostatSettingNumbersNumberCallback(SCtx * ctx, u8 action);
extern u8 ThermostatSettingNumbersNumberDeleteCallback(SCtx * ctx, u8 action);
extern u8 ThermostatSettingNumbersNumberViewCallback (SCtx * ctx, u8 action);
extern u8 ThermostatSettingNumbersNumberAnswersCallback (SCtx * ctx, u8 action);
extern u8 ThermostatSettingNumbersNumberAnswersNumOrToSenderCallback (SCtx * ctx, u8 action);
extern u8 ThermostatSettingNumbersNumberAnswersNumViewCallback (SCtx * ctx, u8 action);
extern u8 ThermostatSettingNumbersNumberAnswersNumChangeCallback (SCtx * ctx, u8 action);
extern u8 ThermostatSettingNumbersNumberAnswersNumDeleteCallback (SCtx * ctx, u8 action);

/* removes all SIM card stored SMSs */
u8 pager_clear_sim_sms (SCtx * ctx, u8 action);

/*

  Thermostat
    Status -> display of the state
    Setting
      On/Off
        On
        Off
      TargetTemp.
      Numbers
        Add
        Tel1
        Tel2
        Tel3
          View
          Delete
          Answers
            To sender's num. / Num
              View
              Change
              Delete (if Num)
*/


/* spider definitions */
SEdgeP g_ThermostatEdges[] = {
  {&g_Thermostat, &g_ThermostatState},
  {&g_Thermostat, &g_ThermostatSetting},
  {&g_ThermostatSetting, &g_ThermostatSettingTurn},
  {&g_ThermostatSetting, &g_ThermostatSettingTemperature},
  {&g_ThermostatSetting, &g_ThermostatSettingNumbers},
  {&g_ThermostatSetting, &g_ThermostatSettingRemoveAllSms},
  {&g_ThermostatSettingNumbers, &g_ThermostatSettingNumbersNew},
  NULL
};

/* main menu */
SNodeP g_Thermostat = { lc_Thermostat, NULL };
SNodeP g_ThermostatState = { lc_State, ThermostatStateCallBack };
SNodeP g_ThermostatSetting = { lc_Setting, NULL };
SNodeP g_ThermostatSettingTurn = { lc_Turn, ThermostatSettingTurnCallBack };
SNodeP g_ThermostatSettingTurnOn = { lc_On, ThermostatSettingTurnOnCallBack };
SNodeP g_ThermostatSettingTurnOff = { lc_Off, ThermostatSettingTurnOffCallBack };
SNodeP g_ThermostatSettingTurnOnWithStar = { lc_OnWithStar, ThermostatSettingTurnOnCallBack };
SNodeP g_ThermostatSettingTurnOffWithStar = { lc_OffWithStar, ThermostatSettingTurnOffCallBack };
SNodeP g_ThermostatSettingTemperature = { lc_Temperature, ThermostatSettingTemperatureCallBack };
SNodeP g_ThermostatSettingNumbers = { lc_Numbers, ThermostatSettingNumbersCallBack };
SNodeP g_ThermostatSettingRemoveAllSms = { lc_Clear_SIM_SMS, pager_clear_sim_sms };
SNodeP g_ThermostatSettingNumbersNew = { LC_NEW, ThermostatSettingNumbersNewCallBack };
SNodeP g_ThermostatSettingNumbersNumberDelete = { LC_DELETE, ThermostatSettingNumbersNumberDeleteCallback };
SNodeP g_ThermostatSettingNumbersNumberView = { LC_VIEW, ThermostatSettingNumbersNumberViewCallback };
SNodeP g_ThermostatSettingNumbersNumberAnswers = { lc_Answers, ThermostatSettingNumbersNumberAnswersCallback };
SNodeP g_ThermostatSettingNumbersNumberAnswersNumView = { LC_VIEW, ThermostatSettingNumbersNumberAnswersNumViewCallback };
SNodeP g_ThermostatSettingNumbersNumberAnswersNumChange = { lc_Change, ThermostatSettingNumbersNumberAnswersNumChangeCallback };
SNodeP g_ThermostatSettingNumbersNumberAnswersNumDelete = { LC_DELETE, ThermostatSettingNumbersNumberAnswersNumDeleteCallback };

/* menu handler */
void ThermostatMenuSelection (Menu_selection_data * x);
/* received SMS handler */
void ThermostatReceiveSMS (File_apdu_data * fd);
/* received call handler */
void ThermostatReceiveCall (u8 * data);

typedef u8 TMsisdnString[ALPHA_LEN];

#define MAX_MSISDN      20

#define MAX_TO_SENDER_STRING_SIZE 12
/* global variable data type */
typedef struct structTPersGlobVar
{
  u8 m_bState;
  b8 m_iExpectedTemperature;
  u8 m_bHeatingOn;
  SEdge *m_AccessControlListHead;
  SNode m_AnswerNumberNode;
  SEdge m_AnswerNumberEdge;
  u8 m_AnswerNumberToSender[MAX_TO_SENDER_STRING_SIZE+1];

#ifdef SMS_WITH_LOGGING
  u8 iMsgCounter;
  TMsisdnString message;
  TMsisdnString sender;
#endif
} TPersGlobVar;

/* pointer to the global variables*/
TPersGlobVar *g_PersGlobVar;

/* temperature times 10 */
typedef b16 TTemperatureTimes10;

/* authorized number */
typedef struct structTAuthorizedNumber {
  u8 *adnAuthorizedNumber; /* the authorized number */
  u8 *adnResponseNumber; /* where to send the responses for the authorized number */
  u8 *adnResponseNumberName; /* where to send the responses for the authorized number */
} TAuthorizedNumber;

/* return whether the thermostat is on/off */
u8 ThermostatGetIsOn(void);
/* init whether the thermostat is on/off */
void ThermostatSetIsOnInit(TPersGlobVar *pPersGlobVar, u8 bSetIsOn);
/* set whether the thermostat is on/off */
void ThermostatSetIsOn(u8 bSetIsOn);

/* return target temperature */
u8 ThermostatGetExpectedTemperature(void);
/* init target temperature */
void ThermostatSetExpectedTemperatureInit(TPersGlobVar *pPersGlobVar, u8 iTemperature);
/* set target temperature */
void ThermostatSetExpectedTemperature(u8 iTemperature);

/* return whether the heating is on */
u8 ThermostatGetIsHeatingOn(void);
/* init whether the heating is on */
void ThermostatSetIsHeatingOnInit(TPersGlobVar *pPersGlobVar, u8 bIsHeatingOn);
/* set whether the heating is on */
void ThermostatSetIsHeatingOn(u8 bIsHeatingOn);
/* turn heating off */
void ThermostatTurnHeatingOff(void);
/* turn heating on */
void ThermostatTurnHeatingOn(void);

/* composing the status for GUI and for SMS */
void ThermostatComposeStatus(u8 *buffer);
/* inits remote numbers */
void ThermostatInitRemoteNumber(TPersGlobVar *pPersGlobVar);
/* inits answer number */
void ThermostatInitAnswerNumber(TPersGlobVar *pPersGlobVar);
/* releases answer number */
void ThermostatReleaseAnswerNumber(TPersGlobVar *pPersGlobVar);
/* releases remote numbers */
void ThermostatReleaseRemoteNumber(TPersGlobVar *pPersGlobVar);
/* add/change the response number */
void ThermostatAddOrChangeResponseNumber(TAuthorizedNumber *pNumberStruct, u8* adnNameInput, u8 *adnMsisdnInput);
/* free response number */
void ThermostatReleaseResponseNumber(TAuthorizedNumber *pNumberStruct);
/* init response number */
void ThermostatInitResponseNumber(TAuthorizedNumber *pNumberStruct);

/* return current temperature time 10  */
TTemperatureTimes10 ThermostatGetRealTemperatureTimes10(void);

#define TEMPERATURE_DEVIATION_TIMES_10 2

/* check the temperature and turn on/off the heating switch as needed */
void ThermostatCheckSwitch(void);
/* handling of received SMS */
void ThermostatProcessSMS(u8 *msg, u8 *strMsisdn);
/* sends responding SMS */
void ThermostatSendResponseSMS(u8* strMsisdn);
/* check if the number is allowed */
u8 ThermostatCheckIsAllowedNumber(u8 *strMsisdn, u8 **pstrMsisdnResponse);

/* pager functions */

/* HW setting/getting */
#define DEF_IN		0
#define DEF_OUT		1
#define DEF_ADC		2

#define DEF_ON		1
#define DEF_OFF		0
/* PIN ids - see the definition of each PINs at the end of the file */
//#define PIN_P13 8
#define PIN_P11 6
#define PIN_P1 9
/* get value at the pin */
u16 get_pin_adc (u8 pin);
/* set what the pin is used for */
void set_dir (u8 pin, u8 val);	// val is DEF_IN, DEF_OUT, DEF_ADC
/* set whether pin is used */
void set_pin (u8 n, u8 val);	//val in DEF_ON, DEF_OFF
/* sending sms */
u8 send_msg (u8 * nr, u8 * msg, u8 * msg1, u8 type);
/* does something with SMSs stored at SIM card */
void refsms_d (File_apdu_data * fa);

/* return whether the thermostat is on/off */
u8 ThermostatGetIsOn(void) {
  return rb ( &g_PersGlobVar->m_bState );
}

/* init whether the thermostat is on/off */
void ThermostatSetIsOnInit(TPersGlobVar *pPersGlobVar, u8 bSetIsOn) {
  wb ( & pPersGlobVar->m_bState, (bSetIsOn) );
}

/* set whether the thermostat is on/off */
void ThermostatSetIsOn(u8 bSetIsOn) {
  u8 bWasOn=ThermostatGetIsOn();
  ThermostatSetIsOnInit(g_PersGlobVar, bSetIsOn);
  /* put in here any possible call backs on changes */
  if ((bWasOn && !bSetIsOn) || (!bWasOn && bSetIsOn)) {
    ThermostatCheckSwitch();
  }
}

/* return target temperature */
u8 ThermostatGetExpectedTemperature(void) {
  return rb ( &g_PersGlobVar->m_iExpectedTemperature );
}
/* init target temperature */
void ThermostatSetExpectedTemperatureInit(TPersGlobVar *pPersGlobVar, u8 iTemperature) {
  wb ( & pPersGlobVar->m_iExpectedTemperature, (iTemperature) );
}
/* set target temperature */
void ThermostatSetExpectedTemperature(u8 iTemperature) {
  u8 iWasTemperature=ThermostatGetExpectedTemperature();
  ThermostatSetExpectedTemperatureInit (g_PersGlobVar, iTemperature);
  /* put in here any possible call backs on changes */
  if (iWasTemperature!=iTemperature) {
    ThermostatCheckSwitch();
  }
}

/* return whether the heating is on */
u8 ThermostatGetIsHeatingOn(void) {
  return rb ( &g_PersGlobVar->m_bHeatingOn );
}
/* init whether the heating is on */
void ThermostatSetIsHeatingOnInit(TPersGlobVar *pPersGlobVar, u8 bIsHeatingOn) {
  wb ( & pPersGlobVar->m_bHeatingOn, (bIsHeatingOn) );
}
/* set whether the heating is on */
void ThermostatSetIsHeatingOn(u8 bIsHeatingOn) {
  u8 bWasHeatingOn=ThermostatGetIsHeatingOn();
  ThermostatSetIsHeatingOnInit (g_PersGlobVar, bIsHeatingOn);
  /* put in here any possible call backs on changes */
  if (bWasHeatingOn && !bIsHeatingOn) {
    /* turn heating off */
    ThermostatTurnHeatingOff();
  } else if (!bWasHeatingOn && bIsHeatingOn) {
    /* turn heating on */
    ThermostatTurnHeatingOn();
  }
}

/* inits answer number */
void ThermostatInitAnswerNumber(TPersGlobVar *pPersGlobVar) {
  SNode *node;
  SEdge *edge;

  memset (& pPersGlobVar->m_AnswerNumberNode, 0, sizeof(SNode) );
  memset (& pPersGlobVar->m_AnswerNumberEdge, 0, sizeof(SEdge) );
  memset (& pPersGlobVar->m_AnswerNumberToSender, 0, MAX_TO_SENDER_STRING_SIZE+1 );

  node = & pPersGlobVar->m_AnswerNumberNode;
  ww (& node->cb, (u16) ThermostatSettingNumbersNumberAnswersNumOrToSenderCallback ); // set the call back

  edge = & pPersGlobVar->m_AnswerNumberEdge;
  ww (&edge->f, (u16) &g_ThermostatSettingNumbersNumberAnswers); // set the pointer to parent to the main access list menu
  ww (&edge->t, (u16) node); // set the pointer to the node
  ww (&edge->next, 0); // prepend the other entries in the access list
  wb (&edge->source, MEM_F_P | MEM_T_E); // state that the from node is from PROGMEM while the to node is from EPROM
}

/* release answer number */
void ThermostatReleaseAnswerNumber(TPersGlobVar *pPersGlobVar) {
}

/* inits remote number */
void ThermostatInitRemoteNumber(TPersGlobVar *pPersGlobVar) {
  ww ( & pPersGlobVar->m_AccessControlListHead, 0 );
}

/* remote item from access control list */
void ThermostatRemoteAccessControlListItem(SEdge *edgeToBeDeleted, SEdge **ppreviousEdge) {
  SNode *currNode;
  SEdge *nextEdge;
  TAuthorizedNumber *pNumberStruct;

  if (edgeToBeDeleted!=NULL) {
    nextEdge = (SEdge *) rw (&edgeToBeDeleted->next); // get the next item
    currNode = (SNode *) rw (&edgeToBeDeleted->t); // get current node
    efree ((u8 *) rw (&currNode->text)); // release the address book name

    // release the number structure
    pNumberStruct = (TAuthorizedNumber *) rw (&currNode->p);
    efree ((u8 *) rw (&pNumberStruct->adnAuthorizedNumber)); // release the authorized number
    ThermostatReleaseResponseNumber(pNumberStruct);
    efree (pNumberStruct); // release the number struct

    efree (currNode); // release the node
    efree (edgeToBeDeleted); // release the item
    
    if (ppreviousEdge!=NULL) {
      ww (ppreviousEdge, (u16) nextEdge); // set the pointer in the previous edge to the next edge
    }
  }
}

/* releases remote numbers */
void ThermostatReleaseRemoteNumber(TPersGlobVar *pPersGlobVar) {

  while ((SEdge *) rw (&pPersGlobVar->m_AccessControlListHead)) { // while the ACL is not empty
    ThermostatRemoteAccessControlListItem( // remove 1st item
      (SEdge *) rw (&pPersGlobVar->m_AccessControlListHead),
      &pPersGlobVar->m_AccessControlListHead
      );
  }
}

/* remove a remote number */
void ThermostatRemoveRemoteNumber(SNode *nodeToBeDeleted) {
  // find the edge with appropriate node
  SEdge *edge = (SEdge *) rw (&g_PersGlobVar->m_AccessControlListHead);
  SEdge **pCurrEdge = &g_PersGlobVar->m_AccessControlListHead;

  while (edge) {
    if ((SNode *) rw (&edge->t) == nodeToBeDeleted) {
      break;
    }
    pCurrEdge = (SEdge **)&edge->next;
    edge = (SEdge *) rw (&edge->next);
  }
  if (edge!=NULL) {
    // remove the found item
    ThermostatRemoteAccessControlListItem(edge,pCurrEdge);
  }
}

/* remove a remote number */
void ThermostatAddRemoteNumber(u8* adnNameInput, u8 *adnMsisdnInput) {
  SNode *node = NULL;
  SEdge *edge = NULL;
  u8 *adnNameCopy = NULL;
  u8 *adnMsisdCopy = NULL;
  u8 i;
  u8 *res;
  TAuthorizedNumber *pNumberStruct=NULL;

  // duplicate the number to eprom

  adnMsisdCopy = msisdncpy (adnMsisdnInput, MSISDN_ADN, MEM_E);
  if (adnMsisdCopy == NULL)
    goto X_END;

  node = emalloc (sizeof (SNode));
  if (node == NULL)
    goto X_END;

  edge = emalloc (sizeof (SEdge));
  if (edge == NULL)
    goto X_END;

  // duplicate the address book name to eprom
  adnNameCopy = emalloc (strlen (adnNameInput) + 1);
  if (adnNameCopy == NULL)
    goto X_END;

  memcpy (adnNameCopy, adnNameInput, strlen (adnNameInput));
  wb (adnNameCopy + strlen (adnNameInput), '\0');

  ww (&node->text, (u16) adnNameCopy); // set the node text to the address bok name
  ww (&node->cb, (u16) ThermostatSettingNumbersNumberCallback); // set the call back to tsms_acl

  // allocate the number struct
  pNumberStruct = emalloc (sizeof(TAuthorizedNumber));
  if (pNumberStruct == NULL)
    goto X_END;

  // allocate the number structure
  ww (&pNumberStruct->adnAuthorizedNumber, (u16) adnMsisdCopy); // keep the authorized number
  ThermostatInitResponseNumber(pNumberStruct);
  ww (&node->p, (u16) pNumberStruct); // set the convenience pointer to the number struct

  ww (&edge->f, (u16) &g_ThermostatSettingNumbers); // set the pointer to parent to the main access list menu
  ww (&edge->t, (u16) node); // set the pointer to the just created node
  ww (&edge->next, (u16) rw (&g_PersGlobVar->m_AccessControlListHead)); // prepend the other entries in the access list
  wb (&edge->source, MEM_F_P | MEM_T_E); // state that the from node is from PROGMEM while the to node is from EPROM

  ww (&g_PersGlobVar->m_AccessControlListHead, (u16) edge); // make the egde a new acess list head
  return;

X_END:
  efree (node);
  efree (edge);
  efree (adnMsisdCopy);
  efree (adnNameCopy);
  return;
}

/* init response number */
void ThermostatInitResponseNumber(TAuthorizedNumber *pNumberStruct) {
  ww (&pNumberStruct->adnResponseNumber, 0);
  ww (&pNumberStruct->adnResponseNumberName, 0);
}

/* free response number */
void ThermostatReleaseResponseNumber(TAuthorizedNumber *pNumberStruct) {
  efree ( (u8 *) rw (&pNumberStruct->adnResponseNumber));
  efree ( (u8 *) rw (&pNumberStruct->adnResponseNumberName));
  ThermostatInitResponseNumber(pNumberStruct);
}

/* add/change the response number */
void ThermostatAddOrChangeResponseNumber(TAuthorizedNumber *pNumberStruct, u8* adnNameInput, u8 *adnMsisdnInput) {
  u8 *adnNameCopy = NULL;
  u8 *adnMsisdCopy = NULL;

  ThermostatReleaseResponseNumber(pNumberStruct);

  adnMsisdCopy = msisdncpy (adnMsisdnInput, MSISDN_ADN, MEM_E);
  if (adnMsisdCopy == NULL)
    goto X_END;

  // duplicate the address book name to eprom
  adnNameCopy = emalloc (strlen (adnNameInput) + 1);
  if (adnNameCopy == NULL)
    goto X_END;

  memcpy (adnNameCopy, adnNameInput, strlen (adnNameInput));
  wb (adnNameCopy + strlen (adnNameInput), '\0');

  ww(&pNumberStruct->adnResponseNumber, (u16) adnMsisdCopy);
  ww(&pNumberStruct->adnResponseNumberName, (u16) adnNameCopy);
  return;

X_END:
  efree (adnMsisdCopy);
  efree (adnNameCopy);
  return;
}

/* main event handles */
void turbo_handler (u8 action, void *data) {
  switch (action) {
    case ACTION_APP_REGISTER: {
      TPersGlobVar *pGlobVars=emalloc (sizeof (TPersGlobVar));
      if (pGlobVars!=NULL) {
        /* init data here
          !!! make sure pGlobVars is used !!!
          !!! global variable g_PersGlobVar is not set yet !!!
          !!! also app_data are not registered yet !!!
          */
        g_PersGlobVar=NULL;
        
        ThermostatSetIsOnInit(pGlobVars, 0);
        ThermostatSetExpectedTemperatureInit(pGlobVars, 5);
        ThermostatSetIsHeatingOnInit(pGlobVars, 0);
        ThermostatInitRemoteNumber(pGlobVars);
        ThermostatInitAnswerNumber(pGlobVars);

#ifdef SMS_WITH_LOGGING
  wb ( & pGlobVars->message[0], 'a' );
  wb ( & pGlobVars->message[1], '\0' );
  wb ( & pGlobVars->sender[0], 'b' );
  wb ( & pGlobVars->sender[1], '\0' );
  wb ( & pGlobVars->iMsgCounter, 0 );
#endif
        /* end of initiation */
        reg_app_data (pGlobVars);
      }
    }
    break;

    case ACTION_APP_UNREGISTER: {
      TPersGlobVar *pGlobVars=app_data ();
      if (pGlobVars!=NULL) {
        ThermostatReleaseRemoteNumber(pGlobVars); // release the access control list
        ThermostatReleaseAnswerNumber(pGlobVars);
      }
      efree (pGlobVars);
    }
    break;

    case ACTION_APP_INIT: {
      g_PersGlobVar=app_data ();
      if (g_PersGlobVar!=NULL) {

        reg_action (ACTION_STATUS);
        reg_action (ACTION_SMS);
        reg_action (ACTION_EVENT_MT_CALL);
      }
    }
    break;

    /* main menu insertion */
    case ACTION_INSERT_MENU: {
      insert_menu (locale (lc_Thermostat));
    }
    break;

    /* menu handling */
    case ACTION_MENU_SELECTION: {
      if (g_PersGlobVar!=NULL) {
        stk_thread (ThermostatMenuSelection, data);
      }
    }
    break;

    /* regular call back */
    case ACTION_STATUS: {
      if (g_PersGlobVar!=NULL) {
        ThermostatCheckSwitch();
      }
    }
    break;

    /* SMS handling */
    case ACTION_SMS: {
      if (g_PersGlobVar!=NULL) {
        stk_thread (ThermostatReceiveSMS, data);
      }
    }
    break;

    case ACTION_EVENT_MT_CALL: {
      stk_thread (ThermostatReceiveCall, data);
    }
    break;
  }
}

/* menu handling */
void ThermostatMenuSelection (Menu_selection_data * x) {
  SCtx *spiderContext;

  spiderContext = spider_init ();
  if (spiderContext != NULL) {
    spider_set_order (spiderContext, ORDER_P, ORDER_R, ORDER_E, ORDER_D);
    spiderContext->eP = &g_ThermostatEdges;
    spiderContext->n = &g_Thermostat;
    spider (spiderContext);
  }
}

/* Thermostat / State callback  */
extern u8 ThermostatStateCallBack(SCtx * ctx, u8 action) {
  if (action == APP_ENTER) {

    u8 *buffer = buf_B ();

    if (buffer!=NULL) {
      u8 retState;
      
      ThermostatComposeStatus(buffer);
      retState = display_text ( buffer, NULL);
      if (retState == APP_END) {
        return APP_END;
      }
    }

    return APP_BACK;
  }
  return APP_OK;
}

/* Thermostat / Setting / Turn callback  */
extern u8 ThermostatSettingTurnCallBack(SCtx * ctx, u8 action) {
  if (action == APP_ENTER) {
    /* select the appropriate strings */
    if (ThermostatGetIsOn()) {
      spider_append_r (ctx, &g_ThermostatSettingTurnOff);
      spider_append_r (ctx, &g_ThermostatSettingTurnOnWithStar);
    } else {
      spider_append_r (ctx, &g_ThermostatSettingTurnOn);
      spider_append_r (ctx, &g_ThermostatSettingTurnOffWithStar);
    }
  } else if (action == APP_LEAVE) {
    spider_clear_r (ctx);
  }
  return APP_OK;
}

/* call back when On is selected */
extern u8 ThermostatSettingTurnOnCallBack (SCtx * ctx, u8 action) {
  if (action == APP_ENTER) {

    ThermostatSetIsOn(1);

    /* go directly to State menu */
    spider_back (ctx);
    return APP_BACK;
  }
  return APP_OK;
}

/* call back when Off is selected */
extern u8 ThermostatSettingTurnOffCallBack (SCtx * ctx, u8 action) {
  if (action == APP_ENTER) {

    ThermostatSetIsOn(0);

    /* go directly to State menu */
    spider_back (ctx);
    return APP_BACK;
  }
  return APP_OK;
}

/* Thermostat / Setting / Temperature callback  */
extern u8 ThermostatSettingTemperatureCallBack (SCtx * ctx, u8 action) {
  if (action == APP_ENTER) {

    u8 expectedTemperaturelabel[10];
    u8 *curPtr = expectedTemperaturelabel;
    u8 *retStr;
    u8 iValue;

    curPtr = sprinti (curPtr, ThermostatGetExpectedTemperature() );
    curPtr = sprintc (curPtr, '\0');

    retStr = get_input (locale (lc_ExpectedTemperature), 1, 2, expectedTemperaturelabel, Q_GET_INPUT_DIGITS);

    if (retStr == ENULL) {
      return APP_END;
    }
    if (retStr == NULL) {
      return APP_BACK;
    }

    retStr [ retStr [0] + 1 ] = '\0';
    iValue = (u8)(atoi (&retStr[2]));
    
    ThermostatSetExpectedTemperature(iValue);

    return APP_BACK;
  }
  return APP_OK;
}

/* changes the buffer -> at least 128 bytes MUST be allocated */
void ThermostatComposeStatus(u8 *buffer) {
  u8 *str = buffer;
  TTemperatureTimes10 iTemperature;

  iTemperature=ThermostatGetRealTemperatureTimes10();

  /* compose the string as folllows
    "Vypnuto, Cilova teptola: XX C, Aktualni teplota YY C";
  */
  

  str = sprints ( str, locale(ThermostatGetIsOn() ? lc_StateOn : lc_StateOff ) );
  str = sprints ( str, locale(lc_TargetTemperature));
  str = sprinti ( str, ThermostatGetExpectedTemperature());
  str = sprints ( str, locale(lc_CurrentTemperature));
  str = sprinti ( str, iTemperature/10);
  str = sprintc ( str, '.');
  str = sprinti ( str, iTemperature%10);
  str = sprints ( str, locale(lc_Celsius));

#ifdef WITH_REMOTE_NUMBERS
// !!!! make sure enough space is allocated
  {
    // go thru the list and destroy individual items
    SEdge *edge = (SEdge *) rw (&g_PersGlobVar->m_AccessControlListHead); // start with head
    SEdge *currEdge;
    SNode *currNode;
    u8 *resMsisdnAdn, *resMsisdnRespAdn;
    u8 *strResMsisdn, *strResMsisdnResp;
    u8 bIsFirst=1;
    TAuthorizedNumber *pNumberStruct;

    str = sprints ( str, locale(lc_RemoteNumbersStart));
    while (edge!=NULL) {
      currEdge = edge; // remember the current item
      edge = (SEdge *) rw (&edge->next); // set iterator to the next one

      currNode = (SNode *) rw (&currEdge->t); // get current node
      
      // read number struct
      pNumberStruct = (TAuthorizedNumber *) rw (&currNode->p); // get the current number structure
      resMsisdnAdn = (u8 *) rw (&pNumberStruct->adnAuthorizedNumber); // get the authorized number
      resMsisdnRespAdn = (u8 *) rw (&pNumberStruct->adnResponseNumber); // get the authorized number

      /* convert to string */
      strResMsisdn=msisdn2str(resMsisdnAdn, MSISDN_ADN, MEM_R);
      
      if (strResMsisdn!=NULL) {
        // prefix with string, if not first
        if (!bIsFirst) {
          str = sprints ( str, locale(lc_RemoteNumbersMiddle));
        }
        bIsFirst=0;
        
        if (resMsisdnRespAdn!=NULL) {
          str = sprints ( str, locale(lc_OpenBracket));
        }
        str = sprints ( str, strResMsisdn);
        if (resMsisdnRespAdn!=NULL) {
          str = sprints ( str, locale(lc_Pointer));
          strResMsisdnResp=msisdn2str(resMsisdnRespAdn, MSISDN_ADN, MEM_R);
          if (strResMsisdnResp!=NULL) {
            str = sprints ( str, strResMsisdnResp);
          }
          free(strResMsisdnResp);
          str = sprints ( str, locale(lc_CloseBracket));
        }

        free(strResMsisdn);
      }
    }
    str = sprints ( str, locale(lc_RemoteNumbersEnd));
  }
#endif

#ifdef SMS_WITH_LOGGING
  str = sprintc ( str, '-');
  str = sprints ( str, g_PersGlobVar->sender);
  str = sprintc ( str, '-');
  str = sprints ( str, g_PersGlobVar->message);
  str = sprintc ( str, '-');
  str = sprinti ( str, rb ( & g_PersGlobVar->iMsgCounter) );
  str = sprintc ( str, '>');
#endif

  str = sprintc ( str, '\0');
}

/* call back on received SMS */
void ThermostatReceiveSMS (File_apdu_data * fd)
{
  u8 *t;
  u8 *data = fd->data;
  u8 i;
  u16 j;
  u8 *s;
  u8 l;
  u8 *strMsisdn = NULL;
  u8 *msg = NULL;


#ifdef SMS_WITH_LOGGING
// count received messages
  wb ( & g_PersGlobVar->iMsgCounter, rb ( & g_PersGlobVar->iMsgCounter ) + 1 );
#endif

  // data points to EFSMS as in 11.11

  // data[0] is Status
  if ((data[0] & 0x07) == 3)	//received, to be read
  {
    s = data; // s points to Status now
    s++;			// s points to length of TS Service Centre Address
    i = *s;   // i is the length of TS Service Centre Address
    s += i + 1;			// s points to the short message TPDU as specified in TS 23.040

    i = *s;  // i is 1st byte of SMS DELIVER type
    if ((i & 0x03) == 0x00)	// TP Message Type Indicator = SMS DELIVER (in the direction SC to MS)
    {
      msg = malloc (161); // allocate enough memory for one SMS text
      if (msg)
      {
      	if (i & 0x40) // TP User Data Header Indicator (TP UDHI) set ?
      	{
      	  // yes, the beginning of the TP UD field contains a Header in addition to the short message.
      	  u16 bits, b1;

      	  dbsp ("UHDI\n");
      	  t = tpdu_seek (s, T_SMS_DCS); // get Data Coding Scheme pointer
      	  j = *t;                      // j is Data Coding Scheme
      	  t = tpdu_seek (s, T_SMS_UDL); // get User Data Length pointer
      	  l = *t;                      // l is the length of the User Data
      	  t++;                         // t points to Information Element Identifier
      	  i = *t;		//UDHL              // i is Information Element Identifier
      	  if (dcs_is_7b (j))
      	  {
      	    dbsp ("7B\n");
      	    j = 7 - ((1 + i) * 8) % 7;	//padding, fill bits
      	    if (j == 7)
      	      j = 0;
      	    dbsp ("JP ");
      	    dbih (j);
      	    dbc ('\n');
      	    j += (1 + i) * 8;	// header in bits divisible 7
      	    dbsp ("J7 ");
      	    dbih (j);
      	    dbc ('\n');
      	    bits = l * 7 - j;
      	    dbsp ("BITS ");
      	    dbih (j);
      	    dbc ('\n');
      	    for (b1 = 0; b1 < bits; b1++)
      	      set_bit (msg, b1, get_bit (t, b1 + j));

      	    i = bits / 7;
      	    dcs_78 (msg, i, DCS_7_TO_8);
      	    msg[i] = '\0';
      	  }
      	  else
      	  {
      	    dbsp ("8B\n");
      	    memcpy (msg, t + 1 + i, l - 1 - i);
      	    msg[l - 1 - i] = '\0';
      	  }
      	}
      	else
      	{
      	  t = tpdu_seek (s, T_SMS_UDL);
      	  i = *t;
      	  memcpy (msg, t + 1, 140);

      	  t = tpdu_seek (s, T_SMS_DCS);
      	  if (dcs_is_7b (*t))
      	    dcs_78 (msg, i, DCS_7_TO_8);
      	  msg[i] = '\0';
      	}

      	t = tpdu_seek (s, T_SMS_AA);
      	strMsisdn = msisdn2str (t, MSISDN_SMS, MEM_R);

        // this seems to pass the SMS to SIM card and then removes it
      	refsms_d (fd);

        /* ms contains the phone number, msg contains the message */

        ThermostatProcessSMS(msg, strMsisdn);

      	free (strMsisdn);
      	free (msg);

      	return;
      }
    }

    // this seems to pass the SMS to SIM card and then removes it
    refsms_d (fd);
    return;
  }

  store_sms_2_sim ();		// not incoming message, something user
}

/* strcmp implemenation replacement (the normal strcmp library function always return 0) */
u8 StrIsSame(u8 *strA, u8 *strB) {
  u16 iASize, iBSize;
  
  iASize=strlen(strA);
  iBSize=strlen(strB);
  if (iASize==iBSize) {
    return memcmp(strA, strB, iASize)==0;
  } else {
    return 0;
  }
}

/* check if the number is allowed */
u8 ThermostatCheckIsAllowedNumber(u8 *strCheckedMsisdn, u8 **pstrMsisdnResponse) {
  // go thru the list and destroy individual items
  SEdge *edge; 
  SEdge *currEdge;
  SNode *currNode;
  u8 *adnMsisdn;
  u8 *adnCheckedMsisdn=NULL;
  u8 bFound=0;
  TAuthorizedNumber *pNumberStruct;

  if (strCheckedMsisdn!=NULL && pstrMsisdnResponse!=NULL) {
    // reset the response number
    *pstrMsisdnResponse=NULL;
    // convert string number to ADN number
    adnCheckedMsisdn=str2msisdn(strCheckedMsisdn, MSISDN_ADN, MEM_R);
    if (adnCheckedMsisdn!=NULL) {
      // go thru the ACL
      edge = (SEdge *) rw (&g_PersGlobVar->m_AccessControlListHead); // start with head

      while (edge!=NULL) {
        currEdge = edge; // remember the current item
        edge = (SEdge *) rw (&edge->next); // set iterator to the next one

        currNode = (SNode *) rw (&currEdge->t); // get current node
        // read number struct
        pNumberStruct = (TAuthorizedNumber *) rw (&currNode->p);
        adnMsisdn = (u8 *) rw (&pNumberStruct->adnAuthorizedNumber); // get the current number

        if (msisdncmp(adnCheckedMsisdn, MSISDN_ADN, adnMsisdn, MSISDN_ADN)) {
          // number found, now create the response number
          
          // create response number
          u8 *adnResponseNum, *strResponseNum;
          adnResponseNum=(u8 *)rw(&pNumberStruct->adnResponseNumber);
          // if no response number defined, take the sender number
          if (adnResponseNum==NULL) {
            adnResponseNum=adnMsisdn;
          }
          // convert ADN number to string num
          strResponseNum=msisdn2str(adnResponseNum, MSISDN_ADN, MEM_R);
          if (strResponseNum!=NULL) {
            *pstrMsisdnResponse=strResponseNum;
            bFound = 1;
          }
          break;
        }
      }
    }
  }
  
  free(adnCheckedMsisdn);
  return bFound;
}

/* handles the received message */
void ThermostatProcessSMS (u8 *msg, u8 *strMsisdn) {
  u8 msgLen;
  u8 iValue;
  u8 bIsOk;
  u8 err;
  u8 *strMsisdnResponse=NULL;

#ifdef SMS_WITH_LOGGING
{
        u16 iNumLen;
        u8 *start;
        u8 *end;

        iNumLen=strlen(strMsisdn);
        iNumLen=iNumLen>sizeof(TMsisdnString)-1?sizeof(TMsisdnString)-1:iNumLen;
        start=g_PersGlobVar->sender;
        end=& g_PersGlobVar->sender[iNumLen];
        /* set output number to the edit output */
        memcpy(start, strMsisdn, iNumLen);
        wb(end, '\0');
}
{
        u16 iNumLen;
        u8 *start;
        u8 *end;

        iNumLen=strlen(msg);
        iNumLen=iNumLen>sizeof(TMsisdnString)-1?sizeof(TMsisdnString)-1:iNumLen;
        start=g_PersGlobVar->message;
        end=& g_PersGlobVar->message[iNumLen];
        /* set output number to the edit output */
        memcpy(start, msg, iNumLen);
        wb(end, '\0');
}
#endif

  /* check if the number is allowed */
  bIsOk=ThermostatCheckIsAllowedNumber(strMsisdn, &strMsisdnResponse);
  /* if the sender is one of our phone numbers */
  if (bIsOk) {
    msgLen=strlen(msg);
    /* check if msg is "ON" */
    if (msgLen>=2 && (msg[0]=='o' || msg[0]=='O') && (msg[1]=='n' || msg[1]=='N')) {
      /* check if msg is "ON <something>" */
      if (msgLen>=4 && msg[2]==' ') {
        /* convert <something> to number */
        iValue = (u8)(atoi(msg+3));
        /* if not 0, the set the temperature */
        if (iValue!=0) {
          ThermostatSetExpectedTemperature(iValue);
        }
      }
      /* in any case, start the thermostat */
      ThermostatSetIsOn(1);
    } else if (msgLen>=3 && (msg[0]=='o' || msg[0]=='O') && (msg[1]=='f' || msg[1]=='F') && (msg[2]=='f' || msg[2]=='F')) {
      /* if OFF */
      /* stop the thermostat */
      ThermostatSetIsOn(0);
    }
    /* send response message */
    ThermostatSendResponseSMS(strMsisdnResponse);
  }
  free(strMsisdnResponse);
}

/* sends response message */
void ThermostatSendResponseSMS(u8* strMsisdn) {
  if (strMsisdn!=NULL) {
    u8 *buffer = buf_B ();
    u8 *msisdnSMS=NULL;
    
    /* convert MSISDN string to SMS address format */
    msisdnSMS=str2msisdn(strMsisdn, MSISDN_SMS, MEM_R);
    if (buffer!=NULL && msisdnSMS!=NULL) {
      ThermostatComposeStatus(buffer);
    	send_msg (msisdnSMS, buffer, NULL, MSISDN_SMS);
    }
  	free(msisdnSMS);
  }
}


/* coversion table for the P1 thermistor resistance, values -5 .. 50 C, step 5 C */
u16 PROGMEM g_ThermistorResistance[] = {
	47077,
  35563,
  27119,
  20860,
  16204,
  12683,
  10000,
  7942,
  6327,
  5074,
  4103,
  3336
};
/* thermistor resistance table len */
#define g_ThermistorResistanceLen (sizeof(g_ThermistorResistance)/sizeof(g_ThermistorResistance[0]))

/* get the current voltage on P1 pin */
u16 ThermostatGetP1Voltage();

/* gets real temperature * 10 */
TTemperatureTimes10 ThermostatGetRealTemperatureTimes10(void) {
  u16 voltage;
  u8 err=NO_ERROR;
  TTemperatureTimes10 iTemperature;

  voltage=ThermostatGetP1Voltage();
  u16 resistance;
  b8 minHigherResistanceIndex, maxLowerResistanceIndex;
  b8 minHigherResistanceTemperature, maxLowerResistanceTemperature;
  u16 minHigherResistanceValue, maxLowerResistanceValue;

  /* calculate K164NK010 thermistor resistance from the 3V VCC and 10k serial resistor */
  resistance=((60000-voltage*50)/voltage)*200;

  // find maximal value lower than resistance
  for(minHigherResistanceIndex=g_ThermistorResistanceLen-2;minHigherResistanceIndex>=0;minHigherResistanceIndex--) {
    if (resistance<rw(&g_ThermistorResistance[minHigherResistanceIndex])) {
      break;
    }
  }
  minHigherResistanceIndex++;
  minHigherResistanceTemperature=(minHigherResistanceIndex-1)*5;
  minHigherResistanceValue=rw(&g_ThermistorResistance[minHigherResistanceIndex]);

  // find minimal value higher than resistance
  for(maxLowerResistanceIndex=1;maxLowerResistanceIndex<g_ThermistorResistanceLen;maxLowerResistanceIndex++) {
    if (resistance>rw(&g_ThermistorResistance[maxLowerResistanceIndex])) {
      break;
    }
  }
  maxLowerResistanceIndex--;
  maxLowerResistanceTemperature=(maxLowerResistanceIndex-1)*5;
  maxLowerResistanceValue=rw(&g_ThermistorResistance[maxLowerResistanceIndex]);

  if (minHigherResistanceIndex!=maxLowerResistanceIndex) {
    // if the termistor does not exactly hit the table value
    iTemperature=
        maxLowerResistanceTemperature*10+
          (b16)
            (
              (
                10*
                ((b32)minHigherResistanceTemperature-(b32)maxLowerResistanceTemperature)*
                ((b32)maxLowerResistanceValue-(b32)resistance)
              )/((b32)maxLowerResistanceValue-(b32)minHigherResistanceValue)
            )
            ;
  } else {
    // if the termistor hit exactly the table value
    iTemperature=maxLowerResistanceTemperature*10;
  }
  return iTemperature;
}

/* gets P1 voltage */
u16 ThermostatGetP1Voltage() {
  set_dir(PIN_P1, DEF_ADC);
  return get_pin_adc(PIN_P1);
}

/* check the temperature and turn on/off the heating switch as needed */
void ThermostatCheckSwitch(void) {

  if ( ThermostatGetIsOn () ) {
    /* if termostat is on */
    TTemperatureTimes10 iTemperature;
    b8 iExpectedTemperature;

    iExpectedTemperature=(b8)ThermostatGetExpectedTemperature();
    iTemperature = ThermostatGetRealTemperatureTimes10();

    /* if temperature is more than 0,7 grade higher than expected and if the heating is still on */
    if (
      (iTemperature > 10*(TTemperatureTimes10)iExpectedTemperature + TEMPERATURE_DEVIATION_TIMES_10)
      && ThermostatGetIsHeatingOn()
      ) {
      /* turn heating off */
      ThermostatSetIsHeatingOn(0);

    } else if ((iTemperature < 10*(TTemperatureTimes10)iExpectedTemperature - TEMPERATURE_DEVIATION_TIMES_10)
      && !ThermostatGetIsHeatingOn()
      ) {
      /*  turn heating on */
      ThermostatSetIsHeatingOn(1);
    }
  } else {
    /* if termostat is off */
    if (ThermostatGetIsHeatingOn()) {
      /* if heating is still running */

      /* turn heating off */
      ThermostatSetIsHeatingOn(0);
    }
  }
}

/* turn heating off */
void ThermostatTurnHeatingOff(void) {
	set_dir (PIN_P11, DEF_OUT);
	set_pin (PIN_P11, DEF_OFF);
}
/* turn heating on */
void ThermostatTurnHeatingOn(void) {
	set_dir (PIN_P11, DEF_OUT);
	set_pin (PIN_P11, DEF_ON);
}


/* taken from pager */

#define CONFIG_ADC_CHANNEL_CHANGE_DELAY_MILISEC (5)
#define ADC_INTERNAL_REF_VOLT (0xC0)

void delay_1ms (void)
{
  u16 i;

// NOTE! we assume clock 3MHz, it can be 2-4Mhz depending on used ME

  for (i = 0; i < 1000; i++)
  {
    __asm__ __volatile__ ("nop");
    __asm__ __volatile__ ("nop");
  }
}


// Measure Voltage on pin of A/D MUX
u16 adc_measure (u8 channel)
{
  u16 msecs;
  u8 i;
  u16 voltage;

  if (channel >= 8)
    return 0;

  //Use Voltage reference as set in init_adc function
  //Select ADC MUX channel
  outb (ADMUX, (ADC_INTERNAL_REF_VOLT | channel));

  for (i = 0; i < 20; i++)
  {
    // wait for a while
    // keep in mind Mr. Nyquist
    // this delay will influence sampling frequency !
    for (msecs = 0; msecs < CONFIG_ADC_CHANNEL_CHANGE_DELAY_MILISEC; msecs++)
      delay_1ms ();

    //Disable high speed ADC mode
    cbi (SFIOR, ADHSM);

    // Enable ADC
    // Start conversion
    // Disable Free running mode
    // ADCclk=Fosc/16
    outb (ADCSR, 0xC4);
    // Wait here when conversion is not completed with data registers updating
    while ((inb (ADCSR) & (1 << ADIF)) == 0);

    //clear ADCIF flag
    sbi (ADCSR, ADIF);

    //store measured voltage
    voltage = inw (ADCW);

    // Disable ADC
    outb (ADCSR, 0x00);
  }

  //Turn Off Internal Voltage reference 2,56V
  outb (ADMUX, 0x00);

  return voltage;
}

/*
PIN TABLE:
0	A	P5
1	B	P6
2	C	P7
3	D	P8
4	E	P9
5	F	P10
6	G	P11
7	H	P12
8	I	P13

Only for Turbo Lite:
9	J	P1
10	K	P2
11	L	P3
12	M	P4

Only for EVO board:
13     N       Opto Out1
14     O       Opto Out2
15     P       Opto In1
16     Q       Opto In2
*/

u16 get_pin_adc (u8 pin)	//pin==9-12
{
  u16 voltage;

  switch (pin)
  {
    case 9:
      voltage = adc_measure (2);
      break;
    case 10:
      voltage = adc_measure (5);
      break;
    case 11:
      voltage = adc_measure (3);
      break;
    case 12:
      voltage = adc_measure (6);
      break;
    default:
      voltage = 0;
      break;
  }
  asm __volatile__ ("nop");	//wait

  return voltage;
}

void set_dir (u8 pin, u8 val)	// val is DEF_IN, DEF_OUT, DEF_ADC
{
  switch (pin)
  {
    case 0:
      if (val == DEF_IN)
        cbi (DDRG, DD1);
      else
        sbi (DDRG, DD1);
      break;
    case 1:
      if (val == DEF_IN)
        cbi (DDRC, DD2);
      else
        sbi (DDRC, DD2);
      break;
    case 2:
      if (val == DEF_IN)
        cbi (DDRC, DD1);
      else
        sbi (DDRC, DD1);
      break;
    case 3:
      if (val == DEF_IN)
        cbi (DDRC, DD6);
      else
        sbi (DDRC, DD6);
      break;
    case 4:
      if (val == DEF_IN)
        cbi (DDRC, DD4);
      else
        sbi (DDRC, DD4);
      break;
    case 5:
      if (val == DEF_IN)
        cbi (DDRA, DD7);
      else
        sbi (DDRA, DD7);
      break;
    case 6:
      if (val == DEF_IN)
        cbi (DDRC, DD7);
      else
        sbi (DDRC, DD7);
      break;
    case 7:
      if (val == DEF_IN)
        cbi (DDRA, DD4);
      else
        sbi (DDRA, DD4);
      break;
    case 8:
      if (val == DEF_IN)
        cbi (DDRA, DD6);
      else
        sbi (DDRA, DD6);
      break;
    case 9:
      if (val == DEF_IN)
        cbi (DDRF, DD2);
      else if (val == DEF_OUT)
        sbi (DDRF, DD2);
      else
      {
        cbi (DDRF, DD2);
        cbi (PORTF, PORT2);
      }
      break;
    case 10:
      if (val == DEF_IN)
        cbi (DDRF, DD5);
      else if (val == DEF_OUT)
        sbi (DDRF, DD5);
      else
      {
        cbi (DDRF, DD5);
        cbi (PORTF, PORT5);
      }
      break;
    case 11:
      if (val == DEF_IN)
        cbi (DDRF, DD3);
      else if (val == DEF_OUT)
        sbi (DDRF, DD3);
      else
      {
        cbi (DDRF, DD3);
        cbi (PORTF, PORT3);
      }
      break;
    case 12:
      if (val == DEF_IN)
        cbi (DDRF, DD6);
      else if (val == DEF_OUT)
        sbi (DDRF, DD6);
      else
      {
        cbi (DDRF, DD6);
        cbi (PORTF, PORT6);
      }
      break;
    case 13:
      if (val == DEF_IN)
        cbi (DDRE, DD2);
      else
        sbi (DDRE, DD2);
      break;
    case 14:
      if (val == DEF_IN)
        cbi (DDRE, DD3);
      else
        sbi (DDRE, DD3);
      break;
    case 15:
      if (val == DEF_IN)
        cbi (DDRE, DD6);
      else
        sbi (DDRE, DD6);
      break;
    case 16:
      if (val == DEF_IN)
        cbi (DDRE, DD7);
      else
        sbi (DDRE, DD7);
      break;
    default:
      break;
  }
  asm __volatile__ ("nop");        //wait
}

void set_pin (u8 n, u8 val)	//val in DEF_ON, DEF_OFF
{
  switch (n)
  {
    case 0:
      if (val != 0)
        sbi (PORTG, PORT1);
      else
        cbi (PORTG, PORT1);
      break;
    case 1:
      if (val != 0)
        sbi (PORTC, PORT2);
      else
        cbi (PORTC, PORT2);
      break;
    case 2:
      if (val != 0)
        sbi (PORTC, PORT1);
      else
        cbi (PORTC, PORT1);
      break;
    case 3:
      if (val != 0)
        sbi (PORTC, PORT6);
      else
        cbi (PORTC, PORT6);
      break;
    case 4:
      if (val != 0)
        sbi (PORTC, PORT4);
      else
        cbi (PORTC, PORT4);
      break;
    case 5:
      if (val != 0)
        sbi (PORTA, PORT7);
      else
        cbi (PORTA, PORT7);
      break;
    case 6:
      if (val != 0)
        sbi (PORTC, PORT7);
      else
        cbi (PORTC, PORT7);
      break;
    case 7:
      if (val != 0)
        sbi (PORTA, PORT4);
      else
        cbi (PORTA, PORT4);
      break;
    case 8:
      if (val != 0)
        sbi (PORTA, PORT6);
      else
        cbi (PORTA, PORT6);
      break;
    case 9:
      if (val != 0)
        sbi (PORTF, PORT2);
      else
        cbi (PORTF, PORT2);
      break;
    case 10:
      if (val != 0)
        sbi (PORTF, PORT5);
      else
        cbi (PORTF, PORT5);
      break;
    case 11:
      if (val != 0)
        sbi (PORTF, PORT3);
      else
        cbi (PORTF, PORT3);
      break;
    case 12:
      if (val != 0)
        sbi (PORTF, PORT6);
      else
        cbi (PORTF, PORT6);
      break;
    case 13:
      if (val != 0)
        sbi (PORTE, PORT2);
      else
        cbi (PORTE, PORT2);
      break;
    case 14:
      if (val != 0)
        sbi (PORTE, PORT3);
      else
        cbi (PORTE, PORT3);
      break;
    case 15:
      if (val != 0)
        sbi (PORTE, PORT6);
      else
        cbi (PORTE, PORT6);
      break;
    case 16:
      if (val != 0)
        sbi (PORTE, PORT7);
      else
        cbi (PORTE, PORT7);
      break;
    default:
      break;
  }

  asm __volatile__ ("nop");        //wait
}

u8 send_msg (u8 * nr, u8 * msg, u8 * msg1, u8 type)
{
  u8 *tmp;
  u8 *r = tmp;
  u8 *s;
  u8 i, j;
  u16 val;
  u8 *ms;

  dbsp (">send_msg\n");
  if (msg == NULL)
    return 0;
  if (nr == NULL)
    return 0;

  tmp = malloc (140);
  if (tmp == NULL)
    return 0;
  ms = msisdncpy (nr, type, MEM_R);
  if (ms == NULL)
  {
    free (tmp);
    return 0;
  }

  i = strlen (msg);
  memcpy (tmp, msg, i);
  if (msg1)
  {
    tmp[i] = '\n';
    j = strlen (msg1);
    if (j + i + 1 > 140)
      j = 140 - i - 1;
    memcpy (tmp + i + 1, msg1, j);
    i = i + j + 1;
  }

  i = dcs_78 (tmp, i, DCS_8_TO_7);
  r = send_sms (tmp, i, ms, type, 0, 0, NULL, NULL);

  free (ms);
  free (tmp);

  i = get_tag (r, T_RESULT);
  if (i != 0)
  {
    i += 2;
    if (r[i] == 0x00)
    {
      dbsp ("SEND_SMS OK\n");
      dbsp ("<send_msg\n");
      return 1;
    }
  }
  dbsp ("SEND_SMS NOT_OK\n");

  dbsp ("<send_msg\n");
  return 0;
}

static u8 PROGMEM ef_sms_path[] = { 0x3F, 0x00, 0x7F, 0x10, 0x6F, 0x3C };

void refsms ()
{
  u8 file_buf[6];

  dbsp (">refresh_sms\n");
  memcpy (file_buf, ef_sms_path, sizeof (ef_sms_path));
  refresh (Q_REFRESH_SIM_FILE_CHANGE, 1, 6, file_buf);
  dbsp ("<refresh_sms\n");
}

void refsms_d (File_apdu_data * fa)
{
  sim (fa->ins, fa->p1, fa->p2, fa->p3, fa->data);
  fa->data[0] = 0;
  sim (fa->ins, fa->p1, fa->p2, fa->p3, fa->data);
  refsms ();
}

u8 pager_clear_sim_sms (SCtx * ctx, u8 action)
{
  if (action == APP_ENTER)
  {
    u16 ret;
    u8 nr_sms;
    u8 rec_len;
    u8 i;
    u8 path[6];

    dbsp ("Clear SIM SMS\n");
    memcpy (path, ef_sms_path, sizeof (ef_sms_path));

    play_tone (locale (lc_Please_Wait), TONE_POSITIVE, DURATION_SEC, 1);
    ret = select (path, 3);
    sim (ME_CMD_GET_RESPONSE, 0, 0, ret, buf_A ());
    nr_sms =
      ((u16) ((u16) (buf_A ()[2]) << 8) | (buf_A ()[3])) / buf_A ()[14];
    rec_len = buf_A ()[14];

    for (i = 0; i < nr_sms; i++)
    {
      dbsp ("DEL SMS ");
      dbch (i + 1);
      dbc ('\n');
      buf_A ()[0] = 0;
      sim (ME_CMD_UPDATE_RECORD, i + 1, 0x04, rec_len, buf_A ());
    }

    select (0, 0);

    refsms ();
    i = display_text (locale (lc_SMS_Deleted), NULL);
    if (i == APP_END)
      return APP_END;
    return APP_BACK;
  }
  return APP_OK;
}

/* copy from tsmsacl */

// menu on each entry -> adds just "delete" to each number
u8 ThermostatSettingNumbersNumberCallback (SCtx * ctx, u8 action)
{
  if (action == APP_ENTER)
  {
    spider_append_r (ctx, &g_ThermostatSettingNumbersNumberDelete);
    spider_append_r (ctx, &g_ThermostatSettingNumbersNumberAnswers);
    spider_append_r (ctx, &g_ThermostatSettingNumbersNumberView);
  }
  else if (action == APP_LEAVE)
  {
    spider_clear_r (ctx);
  }
  return APP_OK;
}

/* call back on new item menu */
u8 ThermostatSettingNumbersNewCallBack (SCtx * ctx, u8 action) {
  if (action == APP_ENTER) {
    u8 alpha[ALPHA_LEN];
    u8 *res;

    // let user insert/select a number (and address book name)

    res = msisdn (alpha);
    if (res == ENULL)
      return APP_END;
    if (res == NULL)
      return APP_BACK;

    /* add remote number */
    ThermostatAddRemoteNumber(alpha, res);
    return APP_BACK;
  }
  return APP_OK;
}

// call back for "delete" on a phone number
u8 ThermostatSettingNumbersNumberDeleteCallback (SCtx * ctx, u8 action) {
  if (action == APP_ENTER) {
    u8 res;

    // get pointer to the number
    SNode *node = ctx->parent->f;

    // ask for confirmation
    res = yesno ((u8 *) rw (&node->text), Q_YESNO_DELETE_QMARK, 0);
    if (res == APP_END)
      return APP_END;
    if (res == APP_YES) {
      ThermostatRemoveRemoteNumber(node);
      spider_back (ctx);
    }
    return APP_BACK;
  }
  return APP_OK;
}

// ACL menu handler
u8 ThermostatSettingNumbersCallBack (SCtx * ctx, u8 action) {
  if (action == APP_ENTER) {
    // initiates the list of eprom edges
    ctx->eE = (SEdge *) rw (&g_PersGlobVar->m_AccessControlListHead);
  }
  else if (action == APP_LEAVE) {
    // removes the list of the eprom edges
    ctx->eE = NULL;
    spider_clear_r (ctx);
  }
  return APP_OK;
}

/* shows the number */
u8 ThermostatSettingNumbersNumberViewCallback (SCtx * ctx, u8 action) {
  if (action == APP_ENTER) {
    u8 *buf = buf_B ();
    u8 *r = buf;
    SNode *np = ctx->parent->f;
    TAuthorizedNumber *pNumberStruct;

    // read numbem struct
    pNumberStruct=(TAuthorizedNumber *) rw (&np->p);
    u8 *s = msisdn2str ( (u8 *) rw (&pNumberStruct->adnAuthorizedNumber), MSISDN_ADN, MEM_R);
    u8 i;

    r = sprints (r, s);
    r = sprintc (r, '\0');
    free (s);
    i = display_text (buf, NULL);
    if (i != APP_END) {
      return APP_BACK;
    }
    return i;
  }
  return APP_OK;
}

/* set up the response number node to either to sender or to a specific number */
void ThermostatSettingNumbersNumberAnswersNode(TAuthorizedNumber *pNumberStruct) {
  u8 *adnResponseNumber;
  u8 *adnResponseNumberName;
  SNode *adnResponseNumberNode;

  adnResponseNumberNode = & g_PersGlobVar->m_AnswerNumberNode;
  adnResponseNumber =(u8 *) rw (&pNumberStruct->adnResponseNumber);
  adnResponseNumberName =(u8 *) rw (&pNumberStruct->adnResponseNumberName);
  ww (& adnResponseNumberNode->p, (u16) pNumberStruct); // set the convenience pointer to the number struct
  if (adnResponseNumber == NULL ) {

    u8 *toSenderText = g_PersGlobVar->m_AnswerNumberToSender;
    u8 *toSenderTextProgMem = locale (lc_ToSender);

    u8 itoSenderTextProgMemLen = strlen(toSenderTextProgMem);
    u8 itoSenderTextProgMemLenMax;

    if (itoSenderTextProgMemLen>MAX_TO_SENDER_STRING_SIZE) {
      itoSenderTextProgMemLenMax=MAX_TO_SENDER_STRING_SIZE;
    } else {
      itoSenderTextProgMemLenMax=itoSenderTextProgMemLen;
    }
    memcpy(toSenderText, toSenderTextProgMem, itoSenderTextProgMemLenMax);
    wb(& toSenderText[itoSenderTextProgMemLenMax],'\0');

/*
    wb(& toSenderText[0],'X');
    wb(& toSenderText[1],'\0');
    {
      u8 *buffer = buf_B ();
      u8 *str=buffer;

      str = sprints ( str, "Testx ");
      str = sprints ( str, toSenderText);
      str = sprintc ( str, '-');
      str = sprints ( str, toSenderTextProgMem);
      str = sprintc ( str, '\0');

      display_text ( buffer, NULL);
    }
*/

    ww (& adnResponseNumberNode->text, (u16) toSenderText); // set the node text to "to sender"
  } else {
    ww (& adnResponseNumberNode->text, (u16) adnResponseNumberName ); // set the node text to the address book name
  }
}

/* shows the answer */
u8 ThermostatSettingNumbersNumberAnswersCallback (SCtx * ctx, u8 action) {
  if (action == APP_ENTER)
  {
    SNode *np = ctx->parent->f;
    TAuthorizedNumber *pNumberStruct;

    // read numbem struct
    pNumberStruct=(TAuthorizedNumber *) rw (&np->p);

    ThermostatSettingNumbersNumberAnswersNode(pNumberStruct);

/*
    {
      u8 *buffer = buf_B ();
      u8 *str=buffer;

      SNode *adnResponseNumberNode =  & g_PersGlobVar->m_AnswerNumberNode;

      str = sprints ( str, "Test1 ");
      str = sprinti ( str, (u16)adnResponseNumberNode);
      str = sprintc ( str, '\0');

      display_text ( buffer, NULL);
    }
*/

    ctx->eE = &g_PersGlobVar->m_AnswerNumberEdge;
  }
  else if (action == APP_LEAVE)
  {
    ctx->eE = NULL;
    spider_clear_r (ctx);
  }
  return APP_OK;
}

/* shows the answer to sender or to number */
u8 ThermostatSettingNumbersNumberAnswersNumOrToSenderCallback (SCtx * ctx, u8 action) {
  if (action == APP_ENTER)
  {
    SNode *adnResponseNumberNode;
    TAuthorizedNumber *pNumberStruct;
    u8 *adnResponseNumber;

    adnResponseNumberNode = ctx->n;

/*
    {
      u8 *buffer = buf_B ();
      u8 *str=buffer;

      str = sprints ( str, "Test2 ");
      str = sprinti ( str, (u16)adnResponseNumberNode);
      str = sprintc ( str, '\0');

      display_text ( buffer, NULL);
    }

*/

    pNumberStruct = (TAuthorizedNumber *)rw (& adnResponseNumberNode->p);
    adnResponseNumber =(u8 *) rw (&pNumberStruct->adnResponseNumber);
    if (adnResponseNumber == NULL ) {
      spider_append_r (ctx, &g_ThermostatSettingNumbersNumberAnswersNumChange);
    } else {
      spider_append_r (ctx, &g_ThermostatSettingNumbersNumberAnswersNumDelete);
      spider_append_r (ctx, &g_ThermostatSettingNumbersNumberAnswersNumChange);
      spider_append_r (ctx, &g_ThermostatSettingNumbersNumberAnswersNumView);
    }
  }
  else if (action == APP_LEAVE)
  {
    spider_clear_r (ctx);
  }
  return APP_OK;
}

/* view number for the answers */
u8 ThermostatSettingNumbersNumberAnswersNumViewCallback (SCtx * ctx, u8 action) {
  if (action == APP_ENTER) {
    u8 *buf = buf_B ();
    u8 *r = buf;
    SNode *np = ctx->parent->f;
    TAuthorizedNumber *pNumberStruct;

    // read numbem struct
    pNumberStruct=(TAuthorizedNumber *) rw (&np->p);
    u8 *s = msisdn2str ( (u8 *) rw (&pNumberStruct->adnResponseNumber), MSISDN_ADN, MEM_R);
    u8 i;

    r = sprints (r, s);
    r = sprintc (r, '\0');
    free (s);
    i = display_text (buf, NULL);
    if (i != APP_END) {
      return APP_BACK;
    }
    return i;
  }
  return APP_OK;
}

/* select a new number for the answers */
u8 ThermostatSettingNumbersNumberAnswersNumChangeCallback (SCtx * ctx, u8 action) {
  if (action == APP_ENTER) {
    u8 alpha[ALPHA_LEN];
    u8 *res;
    TAuthorizedNumber *pNumberStruct;
    SNode *np = ctx->parent->f;

/*

    {
      u8 *buffer = buf_B ();
      u8 *str=buffer;

      str = sprints ( str, "Test3 ");
      str = sprinti ( str, (u16)np);
      str = sprintc ( str, '\0');

      display_text ( buffer, NULL);
    }
    return APP_BACK;

*/
    // let user insert/select a number (and address book name)

    res = msisdn (alpha);
    if (res == ENULL)
      return APP_END;
    if (res == NULL)
      return APP_BACK;

    pNumberStruct=(TAuthorizedNumber *) rw (&np->p);
    // change the number
    ThermostatAddOrChangeResponseNumber(pNumberStruct, alpha, res);
    // reset the answer mode node
    ThermostatSettingNumbersNumberAnswersNode(pNumberStruct);
    spider_back (ctx);
    return APP_BACK;
  }
  return APP_OK;
}

/* reset to answers to sender */
u8 ThermostatSettingNumbersNumberAnswersNumDeleteCallback (SCtx * ctx, u8 action) {
  if (action == APP_ENTER) {
    u8 res;

    // get pointer to the number
    SNode *np = ctx->parent->f;

    // ask for confirmation
    res = yesno ((u8 *) rw (&np->text), Q_YESNO_DELETE_QMARK, 0);
    if (res == APP_END)
      return APP_END;
    if (res == APP_YES) {
      TAuthorizedNumber *pNumberStruct;
      // read numbem struct
      pNumberStruct=(TAuthorizedNumber *) rw (&np->p);
      // release the number
      ThermostatReleaseResponseNumber(pNumberStruct);
      // reset the answer mode node
      ThermostatSettingNumbersNumberAnswersNode(pNumberStruct);
      spider_back (ctx);
    }
    return APP_BACK;
  }
  return APP_OK;
}



/* call back on received call */
void ThermostatReceiveCall (u8 * data)
{
  u8 j, i;
  u8 l;
  u8 ms[MAX_MSISDN];
  u8 err;
  u8 bIsOk;
  u8 *strMsisdnResponse=NULL;

  j = get_tag (data, T_ADDRESS);
  if (j != 0) {
    j++;			//tag
    l = data[j] - 2;

    j++;			//len
    j++;			//ADDRESS_LEN
    j++;			//TON/NPI
    i = 0;

    while (l && i < MAX_MSISDN - 1) {
      if ((data[j] & 0x0F) >= 0 && (data[j] & 0x0F) <= 9)
        ms[i++] = (data[j] & 0x0F) + '0';
      if (((data[j] & 0xF0) >> 4) >= 0 && ((data[j] & 0xF0) >> 4) <= 9)
        ms[i++] = ((data[j] & 0xF0) >> 4) + '0';
      j++;
      l--;
    }

    ms[i] = '\0';

#ifdef SMS_WITH_LOGGING
{
        u16 iNumLen;
        u8 *start;
        u8 *end;

        iNumLen=strlen(ms);
        iNumLen=iNumLen>sizeof(TMsisdnString)-1?sizeof(TMsisdnString)-1:iNumLen;
        start=g_PersGlobVar->sender;
        end=& g_PersGlobVar->sender[iNumLen];
        /* set output number to the edit output */
        memcpy(start, ms, iNumLen);
        wb(end, '\0');
}
      // count received messages
        wb ( & g_PersGlobVar->iMsgCounter, rb ( & g_PersGlobVar->iMsgCounter ) + 1 );
#endif

    /* check if the number is allowed */
    bIsOk=ThermostatCheckIsAllowedNumber(ms, &strMsisdnResponse);

    if (bIsOk) {
      /* send response message */
      ThermostatSendResponseSMS(strMsisdnResponse);
    }
  }
  free(strMsisdnResponse);
}
